/*-
 * Copyright (C)2002..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2002..2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: ini.c,v 1.1 2021/08/01 10:01:38 babolo Exp $"

#define BLIN_COMPAT      3
#define Bpars_COMPAT     3
#define MULAR_COMPAT     0
#define MIFE_COMPAT      5
#define RECOBE_COMPAT    4
#define PGOBLIN_COMPAT   2
#define PGOBLIN_INTERNAL 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <multilar.h>
#include <mife.h>
#include "pgoblin.h"

BLIN_flag pgoblin_default;
static size_t u12[] = {341, 512, 512};

static struct {
    char *nm;
    void *mod;
} mods[PGOBLIN_SYMREGS][3] =
{ { {     NULL, NULL}
  , {     NULL, NULL}
  , {     NULL, NULL}
  }
 ,{ {     NULL, NULL}
  , {     NULL, NULL}
  , {     NULL, NULL}
  }
 ,{ {     NULL, NULL}
  , {     NULL, NULL}
  , {     NULL, NULL}
  }
 ,{ {"0"      , &pgoblin_db_0}
  , {     NULL, NULL}
  , {     NULL, NULL}
  }
 ,{ {"local"  , &pgoblin_job_local}
  , {     NULL, NULL}
  , {     NULL, NULL}
  }
 ,{ {"0"      , &pgoblin_style_0}
  , {     NULL, NULL}
  , {     NULL, NULL}
} };

static void *
/**************************************************************************
 **************************************************************************
 **                                                                      **/
dload(pgoblin_main *options, int domain, int rs, const char *nm)      { /**
 **                                                                      **
 **************************************************************************
 **************************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    char *libname = NULL;
    void *module  = NULL;
    void *handle  = NULL;

    if  (!!options->libexec) {
        asprintf(&libname, "%s/libpgoblin%d%.*s.so", options->libexec, domain, rs, nm);
        if  (!libname) {
            ifBLIN_QW0("No mem");
            goto out;
        }
        if  (!(handle = dlopen(libname, /* RTLD_NOW | */ RTLD_LAZY | RTLD_LOCAL))) {
            ifBLIN_QX1("dlopen %s: %s", libname, dlerror());
    }   }
    if  (!handle) {
        asprintf(&libname, PGOBLIN_LIBDIR"/libpgoblin%d%.*s.so", domain, rs, nm);
        if  (!libname) {
            ifBLIN_QW0("No mem");
            goto out;
        }
        if  (!(handle = dlopen(libname, /* RTLD_NOW | */ RTLD_LAZY | RTLD_LOCAL))) {
            ifBLIN_QX0("dlopen %s: %s", libname, dlerror());
            goto out;
    }   }
    if  ((!!handle) && !(module = dlsym(handle, "pgoblin_module"))) {
        ifBLIN_QX0("dlsym pgoblin_module: %s", dlerror());
        goto out;
    }
out:
    if  (libname) free(libname);
    return(module);
#   undef blin_internal_flags
}

int
/**************************************************************************
 **************************************************************************
 **                                                                      **/
pgoblin_load(pgoblin_main *options, int mode, const char *nm, int rs) { /**
 **                                                                      **
 **************************************************************************
 **************************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    int           loaded = -1;
    int           wrt;
    int           ex;
    int           i;

    if  (0 > rs) rs = strlen(nm);
    ifBLIN_QX4("Load %d [%d]=%.*s", mode, rs, rs, nm);
    switch(mode) {
    case 3:
        for (i = PGOBLIN_MODSHIFT; (i < pgoblin.dbtablesz) && !!pgoblin.dbasetable[i]; ++i) {
            if  (!strncmp(pgoblin.dbasetable[i]->name, nm, rs) && !pgoblin.dbasetable[i]->name[rs]) {
                loaded = i;
                goto out;
        }   }
        if  (i >= pgoblin.dbtablesz) {
            ifBLIN_QX0("No free slots for DB %s", nm);
            goto out;
        }
        if  (PGOBLIN_MODSHIFT == i) {
            pgoblin.dbasetable[PGOBLIN_MODSHIFT] = mods[mode][0].mod;
            if  (!!(wrt = pgoblin_wident(options, i, pgoblin.dbasetable[PGOBLIN_MODSHIFT]->id))) {
                ifBLIN_QW0("pgoblin_wident %d", wrt);
            }
            if  ((ex = pgoblin.dbasetable[PGOBLIN_MODSHIFT]->init( (options->flags & BLIN_MASK)
                                                                 | PGOBLIN_MODSHIFT
                ))                                               ) {
                ifBLIN_QW0("DB %s init failed", pgoblin.dbasetable[PGOBLIN_MODSHIFT]->name);
                pgoblin.dbasetable[PGOBLIN_MODSHIFT] = NULL;
                loaded = (0 > ex) ? ex : -ex;
                goto out;
            }
            ++i;
        }
        if  (!strncmp(mods[mode][0].nm, nm, rs) && !mods[mode][0].nm[rs]) {
            loaded = PGOBLIN_MODSHIFT;
            goto out;
        }
        for (int k = 1; !!mods[mode][k].nm; ++k) {
            if  (!strncmp(mods[mode][k].nm, nm, rs) && !mods[mode][k].nm[rs]) {
                pgoblin.dbasetable[i] = mods[mode][k].mod;
                if  (!!(wrt = pgoblin_wident(options, i, pgoblin.dbasetable[i]->id))) {
                    ifBLIN_QW0("pgoblin_wident %d", wrt);
                }
                if  ((ex = pgoblin.dbasetable[i]->init((options->flags & BLIN_MASK) | i))) {
                    ifBLIN_QW0("DB %s init failed", pgoblin.dbasetable[i]->name);
                    pgoblin.dbasetable[i] = NULL;
                } else {
                    loaded = i;
        }   }   }
        if  (0 > loaded) {
            if  (!!(pgoblin.dbasetable[i] = dload(options, 3, rs, nm))) {
                if  (!!(wrt = pgoblin_wident(options, i, pgoblin.dbasetable[i]->id))) {
                    ifBLIN_QW0("pgoblin_wident %d", wrt);
                }
                if  ((ex = pgoblin.dbasetable[i]->init((options->flags & BLIN_MASK) | i))) {
                    ifBLIN_QW0("DB %s init failed", pgoblin.dbasetable[i]->name);
                    pgoblin.dbasetable[i] = NULL;
                } else {
                    loaded = i;
        }   }   }
        break;
    case 4:
        for (i = PGOBLIN_MODSHIFT; (i < pgoblin.jobtabsiz) && !!pgoblin.jobtable[i]; ++i) {
            if  (!strncmp(pgoblin.jobtable[i]->name, nm, rs) && !pgoblin.jobtable[i]->name[rs]) {
                loaded = i;
                goto out;
        }   }
        if  (i >= pgoblin.jobtabsiz) {
            ifBLIN_QX0("No free slots for JOB %s", nm);
            goto out;
        }
        if  (PGOBLIN_MODSHIFT == i) {
            pgoblin.jobtable[PGOBLIN_MODSHIFT] = mods[mode][0].mod;
            if  (!!(wrt = pgoblin_wident(options, i, pgoblin.jobtable[PGOBLIN_MODSHIFT]->id))) {
                ifBLIN_QW0("pgoblin_wident %d", wrt);
            }
            if  ((ex = pgoblin.jobtable[PGOBLIN_MODSHIFT]->init( (options->flags & BLIN_MASK)
                                                               | PGOBLIN_MODSHIFT
                ))                                             ) {
                ifBLIN_QW0("JOB %s init failed", pgoblin.jobtable[PGOBLIN_MODSHIFT]->name);
                pgoblin.jobtable[PGOBLIN_MODSHIFT] = NULL;
                loaded = (0 > ex) ? ex : -ex;
                goto out;
            } else {
                pgoblin_jobl *j;

                MARK_R_JOB_NGO(0);;
                ;;  options->job[0].pids = mular_create(MULAR_ZERO, 3, sizeof(pgoblin_jobl), u12);
                MARK_R_JOB_NWENT(0);;
                j = mular_add(options->job[0].pids);
                j->cmdn = -1;
                j->pid = getppid();
                j->status = PGOBLIN_NOSTATUS;
                j = mular_add(options->job[0].pids);
                j->cmdn = -1;
                j->pid = getpid();
                j->status = PGOBLIN_NOSTATUS;
            }
            ++i;
        }
        if  (!strncmp(mods[mode][0].nm, nm, rs) && !mods[mode][0].nm[rs]) {
            loaded = PGOBLIN_MODSHIFT;
            goto out;
        }
        for (int k = 1; !!mods[mode][k].nm; ++k) {
            if  (!strncmp(mods[mode][k].nm, nm, rs) && !mods[mode][k].nm[rs]) {
                pgoblin.jobtable[i] = mods[mode][k].mod;
                if  (!!(wrt = pgoblin_wident(options, i, pgoblin.jobtable[i]->id))) {
                    ifBLIN_QW0("pgoblin_wident %d", wrt);
                }
                if  ((ex = pgoblin.jobtable[i]->init((options->flags & BLIN_MASK) | i))) {
                    ifBLIN_QW0("JOB %s init failed", pgoblin.jobtable[i]->name);
                    pgoblin.jobtable[i] = NULL;
                } else {
                    loaded = i;
        }   }   }
        if  (0 > loaded) {
            if  (!!(pgoblin.jobtable[i] = dload(options, 4, rs, nm))) {
                if  (!!(wrt = pgoblin_wident(options, i, pgoblin.jobtable[i]->id))) {
                    ifBLIN_QW0("pgoblin_wident %d", wrt);
                }
                if  ((ex = pgoblin.jobtable[i]->init((options->flags & BLIN_MASK) | i))) {
                    ifBLIN_QW0("DB %s init failed", pgoblin.jobtable[i]->name);
                    pgoblin.jobtable[i] = NULL;
                } else {
                    loaded = i;
        }   }   }
        break;
    case 5:
        for (i = PGOBLIN_MODSHIFT; (i < pgoblin.styletbsz) && !!pgoblin.styletable[i]; ++i) {
            if  (!strncmp(pgoblin.styletable[i]->name, nm, rs) && !pgoblin.styletable[i]->name[rs]) {
                loaded = i;
                goto out;
        }   }
        if  (i >= pgoblin.styletbsz) {
            ifBLIN_QX0("No free slots for DB %s", nm);
            goto out;
        }
        if  (PGOBLIN_MODSHIFT == i) {
            pgoblin.styletable[PGOBLIN_MODSHIFT] = mods[mode][0].mod;
            if  (!!(wrt = pgoblin_wident(options, i, pgoblin.styletable[PGOBLIN_MODSHIFT]->id))) {
                ifBLIN_QW0("pgoblin_wident %d", wrt);
            }
            if  ((ex = pgoblin.styletable[PGOBLIN_MODSHIFT]->init( (options->flags & BLIN_MASK)
                                                                 | PGOBLIN_MODSHIFT
                ))                                               ) {
                ifBLIN_QW0("STYLE %s init failed", pgoblin.styletable[PGOBLIN_MODSHIFT]->name);
                pgoblin.styletable[PGOBLIN_MODSHIFT] = NULL;
                loaded = (0 > ex) ? ex : -ex;
                goto out;
            }
            ++i;
        }
        if  (!strncmp(mods[mode][0].nm, nm, rs) && !mods[mode][0].nm[rs]) {
            loaded = PGOBLIN_MODSHIFT;
            goto out;
        }
        for (int k = 1; !!mods[mode][k].nm; ++k) {
            if  (!strncmp(mods[mode][k].nm, nm, rs) && !mods[mode][k].nm[rs]) {
                pgoblin.styletable[i] = mods[mode][k].mod;
                if  (!!(wrt = pgoblin_wident(options, i, pgoblin.styletable[i]->id))) {
                    ifBLIN_QW0("pgoblin_wident %d", wrt);
                }
                if  ((ex = pgoblin.styletable[i]->init((options->flags & BLIN_MASK) | i))) {
                    ifBLIN_QW0("STYLE %s init failed", pgoblin.styletable[i]->name);
                    pgoblin.styletable[i] = NULL;
                } else {
                    loaded = i;
        }   }   }
        if  (0 > loaded) {
            if  (!!(pgoblin.styletable[i] = dload(options, 5, rs, nm))) {
                if  (!!(wrt = pgoblin_wident(options, i, pgoblin.styletable[i]->id))) {
                    ifBLIN_QW0("pgoblin_wident %d", wrt);
                }
                if  ((ex = pgoblin.styletable[i]->init((options->flags & BLIN_MASK) | i))) {
                    ifBLIN_QW0("DB %s init failed", pgoblin.styletable[i]->name);
                    pgoblin.styletable[i] = NULL;
                } else {
                    loaded = i;
        }   }   }
        break;
    default:;
    }
out:;
    return(loaded);
#   undef blin_internal_flags
}
 
pgoblin_main *
/**********************************************************************
 **********************************************************************
 **                                                                  **/
pgoblin_init(BLIN_flag flags, int outfd) {                          /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
    pgoblin_main *options = NULL;

#   define blin_internal_flags (flags & BLIN_MASK)
    ifBLIN_QX2("+ flags=%08X", flags);
    if  (!(options = calloc(1, sizeof(pgoblin_main)))) {
        ifBLIN_QX0("no mem for options");
        errno = ENOMEM;
    } else {
        pgoblin_options = options;
        options->xtrace = -1;
        options->flags = flags;
        pgoblin_default = flags;
        strncpy(options->vers, VERS, 4);
        options->maxtmout = 3600000;
        strncpy((char*)options->delim, "#", 3);

       MARK_IO_OUT_NGO(0);;
        ;;  IO_RG0.onu = outfd;
        ;;  IO_RG0.wri = (int(*)(int, const void *, size_t))mife_writ;
        ;;  IO_RG0.flags |= PGOBLIN_OUTSET;
        MARK_IO_OUT_NWENT(0);;
    }
    ifBLIN_QX2("- %s", options ? "OK" : "FAIL");
    return(options);
#   undef blin_internal_flags
}

void
/**********************************************************************
 **********************************************************************
 **                                                                  **/
pgoblin_fini(pgoblin_main *options) {                               /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
    int i;

    i = pgoblin.dbtablesz;
    if  (i < pgoblin.jobtabsiz) i = pgoblin.jobtabsiz;
    if  (i < pgoblin.styletbsz) i = pgoblin.styletbsz;
    for (; 0 <= i; --i) {
        if  ((i < pgoblin.styletbsz) && !!pgoblin.styletable[i]) {
            pgoblin.styletable[i]->fini();
        }
        if  ((i < pgoblin.dbtablesz) && !!pgoblin.dbasetable[i]) {
            pgoblin.dbasetable[i]->fini();
        }
        if  ((i < pgoblin.jobtabsiz) && !!pgoblin.jobtable[i]) {
            pgoblin.jobtable[i]->fini();
    }   }
    free(options);
}
